/********************************************************************************
*                                                                               *
*                        G Z S t r e a m   C l a s s e s                        *
*                                                                               *
*********************************************************************************
* Copyright (C) 2002,2003 by Sander Jansen.   All Rights Reserved.              *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: FXGzStream.cpp,v 1.5 2003/04/05 14:22:37 fox Exp $                       *
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXGzStream.h"
#include <stdio.h>

#ifdef HAVE_ZLIB_H
#include "zlib.h"

/*
FXGzFileStream & FXGzMemoryStream
----------------------------------
Based upon FXBzFileStream and FXBzMemoryStream written by Lyle Johnson


Notes
------
  - We probably can do more error handling.
  - All default values. Have to read the documentation to see if we can optimize some stuff.

*/

/*******************************************************************************/

namespace FX {


// Initialize file stream
FXGzFileStream::FXGzFileStream(const FXObject* cont):FXStream(cont){
  gzfile=NULL;
  }


// Save to a file
void FXGzFileStream::saveItems(const void *buf,unsigned long n){
  if(dir!=FXStreamSave){ fxerror("FXGzFileStream::saveItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    if(n>0){
      int err=gzwrite((gzFile*)gzfile,(void*)buf,n);
      if(err==0) code=FXStreamFailure;
      }
    }
  }

// Load from file
void FXGzFileStream::loadItems(void *buf,unsigned long n){
  if(dir!=FXStreamLoad){ fxerror("FXBzFileStream::loadItems: wrong stream direction.\n"); }
  if(code==FXStreamOK){
    if(n>0){
      int err = gzread((gzFile*)gzfile,buf,n);
      if(err==0) code=FXStreamEnd;
      if(err==-1) code=FXStreamFailure;
      }
    }
  }


// Try open file stream
FXbool FXGzFileStream::open(const char* filename,FXStreamDirection save_or_load){

  // Check for NULL filename
  if(!filename){ fxerror("FXGzFileStream::open: NULL filename argument.\n"); }

  // Stream should not yet be open
  if(dir!=FXStreamDead){ fxerror("FXBzFileStream::open: stream is already open.\n"); }

  // Open for read
  if(save_or_load==FXStreamLoad){
    gzfile=gzopen(filename,"rb");
    if(gzfile==NULL){
      code=FXStreamNoRead;
      return FALSE;
      }
    }

  // Open for write
  else{
    gzfile=gzopen(filename,"wb");
    if(gzfile==NULL){
      code=FXStreamNoWrite;
      return FALSE;
      }
    }

  // Do the generic book-keeping
  return FXStream::open(save_or_load);
  }


// Close file stream
FXbool FXGzFileStream::close(){
  if(gzfile){
    gzclose((gzFile*)gzfile);
    gzfile=NULL;
    }
  return FXStream::close();
  }


// Destructor
FXGzFileStream::~FXGzFileStream(){
  if(gzfile){
    gzclose((gzFile*)gzfile);
    gzfile=(void*)-1;
    }
  }

}

#endif
